/*
 * Stupid PCI walking code to determine who generated an NMI
 *
 * Author: Don Zickus <dzickus@redhat.com>
 *
 * License: GPLv2
 */
#include <linux/notifier.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/kdebug.h>

static int unknown_nmi_walk_pci_bus(struct notifier_block *self,
			 unsigned long cmd, void *__args)
{
	struct die_args *args = __args;
	struct pt_regs *regs;
	u8 data;
	struct pci_dev *dev = NULL;
	unsigned char reason;

	switch (cmd) {
	case DIE_NMIUNKNOWN:
		break;

	default:
		return NOTIFY_DONE;
	}

	regs = args->regs;
	reason = args->err;

	/* walk the pci bus to see who flagged themselves with a problem */
	printk("RHNMI: cpu%d - reason 0x%x\n", smp_processor_id(), reason);
	printk("RHNMI: walking the pci tree\n");
	while ((dev = pci_get_device(PCI_ANY_ID, PCI_ANY_ID, dev)) != NULL) {
		pci_read_config_byte(dev,0x7,&data);
		if (data & 0xF9)
			printk("RHNMI: Device %02x:%02x.%x - %x Found\n",dev->bus->number,
				PCI_SLOT(dev->devfn),PCI_FUNC(dev->devfn), data);
	}
	printk("RHNMI: done\n");

	return NOTIFY_DONE;
}

static __read_mostly struct notifier_block unknownnmi_notifier = {
	.notifier_call          = unknown_nmi_walk_pci_bus,
	.next                   = NULL,
	.priority               = 1
};

static struct die_args first_check_args = {
	.regs		= NULL,
	.err		= -1
};

int unknownnmi_init(void)
{
	printk("RHNMI: loaded unknown nmi walk pci bus module\n");
	register_die_notifier(&unknownnmi_notifier);
	unknown_nmi_walk_pci_bus(&unknownnmi_notifier, DIE_NMIUNKNOWN, &first_check_args);
	return 0;
}


void unknownnmi_exit(void)
{
	printk("RHNMI: unloaded unknown nmi walk pci bus module\n");
	unregister_die_notifier(&unknownnmi_notifier);
}
module_init(unknownnmi_init);
module_exit(unknownnmi_exit);

MODULE_LICENSE("GPL");
